# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Reformulations that concern a particular type of constraint.

The reformulations' logic is not found here but defined within the constraint
classes in the form of a :class:`constraint conversion class
<picos.constraints.constraint.ConstraintConversion>`.
"""

import inspect

from .. import constraints
from ..constraints import Constraint, ConstraintConversion
from ..modeling.footprint import Footprint
from .reformulation import Reformulation

# No call to apidoc.api_start: Module defines __all__.
# ----------------------------------------------------


def reformulation_init(self, theObject):
    """Implement :meth:`~.reformulation.Reformulation.__init__`."""
    Reformulation.__init__(self, theObject)

    self.constraintType = self.__class__.CONSTRAINT_TYPE
    self.makeTmpProblem = self.__class__.CONVERSION_TYPE.convert
    self.makeDualValue  = self.__class__.CONVERSION_TYPE.dual


def reformulation_supports(cls, footprint):
    """Implement :meth:`~.reformulation.Reformulation.supports`."""
    return ("con", cls.CONSTRAINT_TYPE) in footprint


def reformulation_predict(cls, footprint):
    """Implement :meth:`~.reformulation.Reformulation.predict`."""
    updates = [("con", cls.CONSTRAINT_TYPE, Footprint.NONE)]

    for subtype, count in footprint[("con", cls.CONSTRAINT_TYPE)].items:
        assert len(subtype) == 1
        subtype = subtype[0]

        for addition in cls.CONVERSION_TYPE.predict(subtype, footprint.options):
            assert isinstance(addition[-1], int)
            updates.append(addition[:-1] + (addition[-1]*count,))

    return footprint.updated(updates)


def reformulation_reform_single(self, constraint, options):
    """Convert a single constraint."""
    assert isinstance(constraint, self.constraintType)

    # Create a temporary problem from the constraint to be replaced.
    tmpProblem = self.makeTmpProblem(constraint, options)

    # Keep track of auxilary vars/cons replacing the constraint.
    self.auxVars[constraint] = {}
    self.auxCons[constraint] = []

    # If the constraint to be transformed is part of the output prolem, remove
    # it. This is the case when forwarding but not when updating.
    if constraint.id in self.output.constraints:
        self.output.remove_constraint(constraint.id)

    # Remember auxiliary variables added so that their value can be removed from
    # the solution in reformulation_backward.
    for tmpVarName, tmpVar in tmpProblem.variables.items():
        self.auxVars[constraint][tmpVarName] = tmpVar

    # Add auxiliary constraints to the output problem.
    # HACK: This only works while Problem.constraints is an OrderedDict as
    #       ConstraintConversion needs a way to identify the constraints added
    #       to the temporary problem without having any pointer to them.
    auxCons = self.output.add_list_of_constraints(
        tmpProblem.constraints.values())
    self.auxCons[constraint].extend(auxCons)


def reformulation_forward(self):
    """Implement :meth:`~.reformulation.Reformulation.forward`."""
    self.output = self.input.clone(copyOptions=False)

    self.auxVars = {}
    self.auxCons = {}

    # TODO: Give Problem quick iterators over constraints of certain type.
    for constraint in self.input.constraints.values():
        if isinstance(constraint, self.constraintType):
            self._reform_single(constraint, self.input.options)


def reformulation_update(self):
    """Implement :meth:`~.reformulation.Reformulation.update`."""
    # Pass changes in the objective function.
    self._pass_updated_objective()

    # Pass all variables as they are.
    self._pass_updated_vars()

    # Pass all unhandled constraints as they are.
    added, removed = self._pass_updated_cons(ignore=self.constraintType)

    # Pass all option changes.
    self._pass_updated_options()

    # Reformulate new relevant constraints.
    for constraint in added:
        self._reform_single(constraint, self.input.options)

    # Remove auxiliary objects added for relevant removed constraints.
    for constraint in removed:
        assert constraint in self.auxVars and constraint in self.auxCons

        for auxCon in self.auxCons[constraint]:
            assert auxCon.id in self.output.constraints
            self.output.remove_constraint(auxCon.id)

        self.auxCons.pop(constraint)
        self.auxVars.pop(constraint)


def reformulation_backward(self, solution):
    """Implement :meth:`~.reformulation.Reformulation.backward`."""
    # TODO: Give Problem quick iterators over constraints of certain type.
    for constraint in self.input.constraints.values():
        if isinstance(constraint, self.constraintType):
            try:
                solution.duals[constraint] = self.makeDualValue(
                    {v: solution.primals.get(v)
                        for v in self.auxVars[constraint]},
                    [solution.duals.get(c) for c in self.auxCons[constraint]],
                    self.input.options)
            except NotImplementedError:
                solution.duals[constraint] = None

            for auxVar in self.auxVars[constraint]:
                if auxVar in solution.primals:
                    solution.primals.pop(auxVar)

            for auxCon in self.auxCons[constraint]:
                if auxCon in solution.duals:
                    solution.duals.pop(auxCon)

    return solution


def make_constraint_reformulation(constraint, conversion):
    """Produce a :class:`Reformulation` from a :class:`ConstraintConversion`.

    A helper that creates a :class:`Reformulation` type (subclass) from a
    :class:`Constraint` and :class:`constraint.ConstraintConversion` types.
    """
    assert constraint.__name__.endswith("Constraint"), \
        "Constraint types must have a name ending in 'Constraint'."

    assert conversion.__name__.endswith("Conversion"), \
        "Constraint conversions must have a name ending in 'Conversion'."

    constraintName = constraint.__name__[:-len("Constraint")]
    conversionName = conversion.__name__[:-len("Conversion")]

    name = "{}{}Reformulation".format(constraintName,
        "To{}".format(conversionName) if conversionName else "")

    docstring = "Reformulation created from :class:`{0}.{1} " \
        "<picos.constraints.{0}.{1}>`.".format(
        constraint.__name__, conversion.__name__)

    # TODO: This would need an additional prediction that yields all class
    #       types of constraints that can be converted. Is that needed?
    # NOTE: SumExponentialsConstraint.LSEConversion makes use of return the
    #       constraint as is for certain unsupported subtypes.
    # assert constraint not in conversion.adds_constraint_types(), \
    #     "Constraint conversions may not add the very type being converted."

    body = {
        # Class constants.
        "CONSTRAINT_TYPE": constraint,
        "CONVERSION_TYPE": conversion,

        # Class methods.
        "supports": classmethod(reformulation_supports),
        "predict":  classmethod(reformulation_predict),

        # Instance methods.
        "__init__":       reformulation_init,
        "__doc__":        docstring,
        "_reform_single": reformulation_reform_single,
        "forward":        reformulation_forward,
        "update":         reformulation_update,
        "backward":       reformulation_backward
    }

    # TODO: Check if anything like the following is still necessary.
    # # HACK: See QuadConstraint.ConicConversion.predict.
    # if constraint is constraints.QuadConstraint \
    # and conversion is constraints.QuadConstraint.ConicConversion:
    #     body["_verify_prediction"] = lambda self: None

    return type(name, (Reformulation,), body)


# Allow __init__ to import exactly the generated reformulations using asterisk.
__all__ = []

# For every constraint conversion, generate a problem reformulation.
NUM_REFORMS = 0
CONSTRAINT_TO_REFORMS = {}
for constraint in constraints.__dict__.values():
    if not inspect.isclass(constraint):
        continue

    if not issubclass(constraint, Constraint):
        continue

    CONSTRAINT_TO_REFORMS[constraint] = []

    for conversion in constraint.__dict__.values():
        if not inspect.isclass(conversion):
            continue

        if not issubclass(conversion, ConstraintConversion):
            continue

        reformulation = make_constraint_reformulation(constraint, conversion)

        NUM_REFORMS += 1
        CONSTRAINT_TO_REFORMS[constraint].append(reformulation)

        # Export the reformulations as if they were defined at module level.
        globals()[reformulation.__name__] = reformulation
        __all__.append(reformulation.__name__)

# Make the order of __all__ deterministic.
__all__ = sorted(__all__)

# FIXME: Restore a topological sorting.
# TODO: As above, this would need a separate predictor.
TOPOSORTED_REFORMS = [globals()[name] for name in __all__]


# --------------------------------------------------
# No call to apidoc.ape_end: Module defines __all__.



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`SumExtremesConstraint`."""

from collections import namedtuple

import cvxopt as cvx

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class SumExtremesConstraint(Constraint):
    """Bound on a sum over extreme (eigen)values."""

    class Conversion(ConstraintConversion):
        """Sum over extremes to LMI/affine constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import (HermitianVariable, RealVariable,
                                       SymmetricVariable)
            from . import AffineConstraint, ComplexLMIConstraint, LMIConstraint

            nm, k, eigenvalues, complex = subtype

            # Determine matrix variable dimension.
            if eigenvalues:
                nFloat = nm**0.5
                n = int(nFloat)
                assert n == nFloat
                d = n**2 if complex else n*(n + 1) // 2
            else:
                n = nm

            # Validate k.
            assert k > 0 and k <= n

            # Define shorthands for better formatting below.
            RV, AC = RealVariable, AffineConstraint
            LMI = ComplexLMIConstraint if complex else LMIConstraint
            MV = HermitianVariable if complex else SymmetricVariable

            if eigenvalues:
                if k == 1:
                    yield ("con", LMI.make_type(diag=n), 1)
                elif k == n:
                    # NOTE: Refinement prevents this case from happening.
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
                else:
                    yield ("var", RV.make_var_type(dim=1, bnd=0), 1)
                    yield ("var", MV.make_var_type(dim=d, bnd=0), 1)
                    yield ("con", LMI.make_type(diag=n), 2)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
            else:
                if k == 1:
                    yield ("con", AC.make_type(dim=n, eq=False), 1)
                elif k == n:
                    # NOTE: Refinement prevents this case from happening.
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
                else:
                    yield ("var", RV.make_var_type(dim=1, bnd=0), 1)
                    yield ("var", RV.make_var_type(dim=n, bnd=n), 1)
                    yield ("con", AC.make_type(dim=n, eq=False), 1)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import (Constant, HermitianVariable,
                                       RealVariable, SymmetricVariable)
            from ..modeling import Problem

            theSum   = con.theSum
            relation = con.relation
            rhs      = con.rhs

            x = theSum.x
            k = theSum.k

            if theSum.eigenvalues:
                n = x.shape[0]
                I = Constant('I', cvx.spdiag([1.] * n))
            else:
                n = len(x)

            if x.complex:
                MatrixVariable = HermitianVariable
            else:
                MatrixVariable = SymmetricVariable

            P = Problem()

            if relation == Constraint.LE:
                if theSum.eigenvalues:
                    if k == 1:
                        P.add_constraint(x << rhs * I)
                    elif k == n:
                        # NOTE: Refinement prevents this case from happening.
                        P.add_constraint(("I" | x) <= rhs)
                    else:
                        s = RealVariable('s')
                        Z = MatrixVariable('Z', n)
                        P.add_constraint(Z >> 0)
                        P.add_constraint(x << Z + s * I)
                        P.add_constraint(rhs >= (I | Z) + (k * s))
                else:
                    if k == 1:
                        P.add_constraint(x <= rhs)
                    elif k == n:
                        P.add_constraint((1 | x) <= rhs)
                    else:
                        lbda = RealVariable('lambda')
                        mu = RealVariable('mu', x.shape, lower=0)
                        P.add_constraint(x <= lbda + mu)
                        P.add_constraint(k * lbda + (1 | mu) <= rhs)
            else:
                if theSum.eigenvalues:
                    if k == 1:
                        P.add_constraint(x >> rhs * I)
                    elif k == n:
                        # NOTE: Refinement prevents this case from happening.
                        P.add_constraint((I | x) <= rhs)
                    else:
                        s = RealVariable('s')
                        Z = MatrixVariable('Z', n)
                        P.add_constraint(Z >> 0)
                        P.add_constraint(-x << Z + s * I)
                        P.add_constraint(-rhs >= (I | Z) + (k * s))
                else:
                    if k == 1:
                        P.add_constraint(x >= rhs)
                    elif k == n:
                        P.add_constraint((1 | x) >= rhs)
                    else:
                        lbda = RealVariable('lambda')
                        mu = RealVariable('mu', x.shape, lower=0)
                        P.add_constraint(-x <= lbda + mu)
                        P.add_constraint(k * lbda + (1 | mu) <= -rhs)

            return P

    def __init__(self, theSum, relation, rhs):
        """Construct a :class:`SumExtremesConstraint`.

        :param ~picos.expressions.SumExtremes theSum:
            Left hand side expression.
        :param str relation:
            Constraint relation symbol.
        :param ~picos.expressions.AffineExpression rhs:
            Right hand side expression.
        """
        from ..expressions import AffineExpression, SumExtremes

        assert isinstance(theSum, SumExtremes)
        assert isinstance(rhs, AffineExpression)
        assert relation in self.LE + self.GE
        assert len(rhs) == 1

        self.theSum   = theSum
        self.relation = relation
        self.rhs      = rhs

        super(SumExtremesConstraint, self).__init__(theSum._typeStr)

    Subtype = namedtuple("Subtype", ("argdim", "k", "eigenvalues", "complex"))

    def _subtype(self):
        return self.Subtype(len(self.theSum.x), self.theSum.k,
            self.theSum.eigenvalues, self.theSum.x.complex)

    @classmethod
    def _cost(cls, subtype):
        nm, _, eigenvalues, _ = subtype

        if eigenvalues:
            nFloat = nm**0.5
            n = int(nFloat)
            assert n == nFloat
        else:
            n = nm

        return n + 1

    def _expression_names(self):
        yield "theSum"
        yield "rhs"

    def _str(self):
        if self.relation == self.LE:
            return glyphs.le(self.theSum.string, self.rhs.string)
        else:
            return glyphs.ge(self.theSum.string, self.rhs.string)

    def _get_slack(self):
        if self.relation == self.LE:
            return self.rhs.value - self.theSum.value
        else:
            return self.theSum.value - self.rhs.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements all mathematical variable types."""

import random
import threading
from collections import namedtuple
from copy import copy

import cvxopt

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from ..containers import DetailedType
from .data import cvxopt_equals, cvxopt_maxdiff, load_data, load_shape
from .exp_affine import AffineExpression, ComplexAffineExpression
from .expression import Expression, NotValued
from .vectorizations import (ComplexVectorization, FullVectorization,
                             HermitianVectorization,
                             LowerTriangularVectorization,
                             SkewSymmetricVectorization,
                             SymmetricVectorization,
                             UpperTriangularVectorization)

_API_START = api_start(globals())
# -------------------------------


# The variable IDs start at a random value to prevent a clash if variables are
# pickled and loaded in another python session. The choice of 31 bits is to
# produce up to 2**31-1 short hexadecimal IDs for use in private variable names.
_NEXT_VARIABLE_ID = int(random.getrandbits(31))

# A lock for _NEXT_VARIABLE_ID, if the user uses threads to create variables.
_VARIABLE_ID_LOCK = threading.Lock()


def _make_variable_id(dim):
    """Create a unique (starting) ID for a new variable."""
    with _VARIABLE_ID_LOCK:
        global _NEXT_VARIABLE_ID
        id = _NEXT_VARIABLE_ID
        _NEXT_VARIABLE_ID += dim
        return id


# A map used when converting IDs to strings.
_ID_HEX2STR_MAP = {ord(a): ord(b) for a, b in zip(
    "0123456789abcdef", "ABCDEFGHIJKLMNOP")}
_ID_HEX2STR_MAP_PY2 = "".join([chr(_ID_HEX2STR_MAP[i]) if i in _ID_HEX2STR_MAP
    else "_" for i in range(256)])


def _id_to_str(id):
    """Convert a variable id to a short string identifier."""
    try:
        return "{:08x}".format(id).translate(_ID_HEX2STR_MAP)
    except TypeError:  # Python 2
        return "{:08x}".format(id).translate(_ID_HEX2STR_MAP_PY2)


#: Absolute integrality tolerance for integral and binary variables.
INTEGRALITY_ABS_TOL = 1e-4


class VariableType(DetailedType):
    """The detailed type of a variable for predicting reformulation outcomes."""

    pass


class BaseVariable:
    """Primary base class for all variable types.

    Variables need to inherit this class with priority (first class listed) and
    :class:`~.exp_affine.ComplexAffineExpression` or
    :class:`~.exp_affine.AffineExpression` without priority.
    """

    # TODO: Document changed variable bound behavior: Only full bounds can be
    #       given but they may contain (-)float("inf").
    def __init__(self, name, vectorization, lower=None, upper=None):
        """Perform basic initialization for :class:`BaseVariable` instances.

        :param str name:
            Name of the variable. A leading `"__"` denotes a private variable
            and is replaced by a sequence containing the variable's unique ID.

        :param vectorization:
            Vectorization format used to store the value.
        :type vectorization:
            ~picos.expressions.vectorizations.BaseVectorization

        :param lower:
            Constant lower bound on the variable. May contain ``float("-inf")``
            to denote unbounded elements.

        :param upper:
            Constant upper bound on the variable. May contain ``float("inf")``
            to denote unbounded elements.
        """
        if not isinstance(self, Expression):
            raise TypeError("{} may not be initialized directly.".format(
                type(self).__name__))

        self._id    = _make_variable_id(vectorization.dim)
        self._vec   = vectorization
        self._lower = None if lower is None else self._load_vectorized(lower)
        self._upper = None if upper is None else self._load_vectorized(upper)
        self._value = None

        if not name or name.startswith("__"):
            id_str = "__{}_".format(_id_to_str(self._id))
            name = name.replace("__", id_str, 1) if name else id_str + "_"

        self._name = name

    @property
    def id(self):
        """The unique (starting) ID of the variable, assigned at creation."""
        return self._id

    VarSubtype = namedtuple("VarSubtype", ("dim", "bnd"))

    @classmethod
    def make_var_type(cls, *args, **kwargs):
        """Create a detailed variable type from subtype parameters.

        See also :attr:`var_type`.
        """
        return VariableType(cls, cls.VarSubtype(*args, **kwargs))

    @property
    def var_subtype(self):
        """The subtype part of the detailed variable type.

        See also :attr:`var_type`.
        """
        return self.VarSubtype(self.dim, self.num_bounds)

    @property
    def var_type(self):
        """The detailed variable type.

        This intentionally does not override
        :meth:`Expression.type <.expression.Expression.type>` so that the
        variable still behaves as the affine expression that it represents when
        prediction constraint outcomes.
        """
        return VariableType(self.__class__, self.var_subtype)

    def id_at(self, index):
        """Return the unique ID of a scalar entry, assigned at creation."""
        if index < 0 or index >= self._vec.dim:
            raise IndexError("Bad index {} for the {} dimensional variable {}."
                .format(index, self._vec.dim, self.name))

        return self._id + index

    @property
    def dim(self):
        """The variable's dimension on the real field.

        This corresponds to the length of its vectorized value.
        """
        return self._vec.dim

    @property
    def name(self):
        """The name of the variable."""
        return self._name

    @cached_property
    def long_string(self):
        """Long string representation for printing a :meth:`~picos.Problem`."""
        lower, upper = self.bound_dicts
        if lower and upper:
            bound_str = "clamped"
        elif lower:
            bound_str = "bounded below"
        elif upper:
            bound_str = "bounded above"
        else:
            bound_str = None

        return "{} {} {}{}".format(
            glyphs.shape(self.shape),
            self._get_type_string_base().lower(),
            self.name,
            " ({})".format(bound_str) if bound_str else "")

        return "{} {} variable {}{}".format(
            self._get_type_string_base().rsplit(" Variable")[0].lower(),
            glyphs.shape(self.shape),
            self.name,
            " ({})".format(bound_str) if bound_str else "")

    @cached_property
    def bound_dicts(self):
        """Variable bounds as a pair of mappings from index to scalar bound.

        The indices and bound values are with respect to the internal
        representation of the variable, whose value can be accessed with
        :attr:`internal_value`.

        Upper and lower bounds set to ``float("inf")`` and ``float("-inf")``
        on variable creation, respectively, are not included.
        """
        posinf = float("+inf")
        neginf = float("-inf")

        if self._lower is None:
            lower = {}
        else:
            lower = {i: self._lower[i] for i in range(self.dim)
                     if self._lower[i] != neginf}

        if self._upper is None:
            upper = {}
        else:
            upper = {i: self._upper[i] for i in range(self.dim)
                     if self._upper[i] != posinf}

        return (lower, upper)

    @property
    def num_bounds(self):
        """Number of scalar bounds associated with the variable."""
        lower, upper = self.bound_dicts
        return len(lower) + len(upper)

    @cached_property
    def bound_constraint(self):
        """The variable bounds as a PICOS constraint, or :obj:`None`."""
        lower, upper = self.bound_dicts

        I, J, V, b = [], [], [], []

        for i, bound in upper.items():
            I.append(i)
            J.append(i)
            V.append(1.0)
            b.append(bound)

        offset = len(I)

        for i, bound in lower.items():
            I.append(offset + i)
            J.append(i)
            V.append(-1.0)
            b.append(-bound)

        if not I:
            return None

        A = cvxopt.spmatrix(V, I, J, size=(len(I), self.dim), tc="d")

        Ax = AffineExpression(string=glyphs.Fn("bnd_con_lhs")(self.name),
            shape=len(I), coefficients={self: A})

        return Ax <= b

    def copy(self, new_name=None):
        """Return an independent copy of the variable."""
        name = self.name if new_name is None else new_name

        if self._lower is not None or self._upper is not None:
            return self.__class__(name, self.shape, self._lower, self._upper)
        else:
            return self.__class__(name, self.shape)

    def _load_vectorized(self, value):
        """Support :meth:`__init__` and ``_set_value``."""
        return self._vec.vectorize(
            load_data(value, self._vec.shape, self._typecode)[0])

    def _check_internal_value(self, value):
        """Support :meth:`_set_value` and :meth:`_set_internal_value`."""
        pass

    # NOTE: This needs to be inherited with priority.
    def _get_value(self):
        return self._vec.devectorize(self._get_internal_value())

    # NOTE: This needs to be inherited with priority.
    def _set_value(self, value):
        if value is None:
            self._value = None
            return

        try:
            value = self._load_vectorized(value)
            self._check_internal_value(value)
        except Exception as error:
            raise type(error)("Failed to assign a value to variable {}: {}"
                .format(self.string, error))
        else:
            self._value = value

    def _get_internal_value(self):
        if self._value is None:
            raise NotValued("Variable {} is not valued.".format(self.string))
        else:
            return copy(self._value)

    def _set_internal_value(self, value):
        if value is None:
            self._value = None
            return

        try:
            value = load_data(value, (self._vec.dim, 1), "d")[0]
            self._check_internal_value(value)
        except Exception as error:
            raise type(error)(
                "Failed to assign an internal value to variable {}: {}"
                .format(self.string, error))
        else:
            self._value = value

    internal_value = property(
        lambda self: self._get_internal_value(),
        lambda self, value: self._set_internal_value(value),
        lambda self: self._set_internal_value(None),
        """The internal (special vectorized) value of the variable.""")

    def _get_clstype(self):
        # Mimic an ordinary (complex) affine expression.
        return self._basetype

    def _get_subtype(self):
        # Mimic an ordinary (complex) affine expression.
        return self._basetype._get_subtype(self)

    @classmethod
    def _predict(cls, subtype, relation, other):
        # Mimic an ordinary (complex) affine expression.
        return cls._get_basetype()._predict(subtype, relation, other)


class RealVariable(BaseVariable, AffineExpression):
    """A real-valued variable."""

    def __init__(self, name, shape=(1, 1), lower=None, upper=None):
        """Create a :class:`RealVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of a vector or matrix variable.
        :type shape: int or tuple or list
        :param lower: Constant lower bound on the variable. May contain
            ``float("-inf")`` to denote unbounded elements.
        :param upper: Constant upper bound on the variable. May contain
            ``float("inf")`` to denote unbounded elements.
        """
        shape = load_shape(shape)
        vec = FullVectorization(shape)
        BaseVariable.__init__(self, name, vec, lower, upper)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Real Variable"


class IntegerVariable(BaseVariable, AffineExpression):
    """An integer-valued variable."""

    def __init__(self, name, shape=(1, 1), lower=None, upper=None):
        """Create an :class:`IntegerVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of a vector or matrix variable.
        :type shape: int or tuple or list
        :param lower: Constant lower bound on the variable. May contain
            ``float("-inf")`` to denote unbounded elements.
        :param upper: Constant upper bound on the variable. May contain
            ``float("inf")`` to denote unbounded elements.
        """
        shape = load_shape(shape)
        vec = FullVectorization(shape)
        BaseVariable.__init__(self, name, vec, lower, upper)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Integer Variable"

    def _check_internal_value(self, value):
        fltData = list(value)

        if not fltData:
            return  # All elements are exactly zero.

        intData = cvxopt.matrix([round(x) for x in fltData])
        fltData = cvxopt.matrix(fltData)

        if not cvxopt_equals(intData, fltData, INTEGRALITY_ABS_TOL):
            raise ValueError("Data is not near-integral with absolute tolerance"
                " {:.1e}: Largest difference is {:.1e}.".format(
                INTEGRALITY_ABS_TOL, cvxopt_maxdiff(intData, fltData)))


class BinaryVariable(BaseVariable, AffineExpression):
    r"""A :math:`\{0,1\}`-valued variable."""

    def __init__(self, name, shape=(1, 1)):
        """Create a :class:`BinaryVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of a vector or matrix variable.
        :type shape: int or tuple or list
        """
        shape = load_shape(shape)
        vec = FullVectorization(shape)
        BaseVariable.__init__(self, name, vec)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Binary Variable"

    def _check_internal_value(self, value):
        fltData = list(value)

        if not fltData:
            return  # All elements are exactly zero.

        binData = cvxopt.matrix([float(bool(round(x))) for x in fltData])
        fltData = cvxopt.matrix(fltData)

        if not cvxopt_equals(binData, fltData, INTEGRALITY_ABS_TOL):
            raise ValueError("Data is not near-binary with absolute tolerance"
                " {:.1e}: Largest difference is {:.1e}.".format(
                INTEGRALITY_ABS_TOL, cvxopt_maxdiff(binData, fltData)))


class ComplexVariable(BaseVariable, ComplexAffineExpression):
    """A complex-valued variable.

    Passed to solvers as a real variable vector with :math:`2mn` entries.
    """

    def __init__(self, name, shape=(1, 1)):
        """Create a :class:`ComplexVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of a vector or matrix variable.
        :type shape: int or tuple or list
        """
        shape = load_shape(shape)
        vec = ComplexVectorization(shape)
        BaseVariable.__init__(self, name, vec)
        ComplexAffineExpression.__init__(
            self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Complex Variable"


class SymmetricVariable(BaseVariable, AffineExpression):
    r"""A symmetric matrix variable.

    Stored internally and passed to solvers as a symmetric vectorization with
    only :math:`\frac{n(n+1)}{2}` entries.
    """

    def __init__(self, name, shape=(1, 1), lower=None, upper=None):
        """Create a :class:`SymmetricVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of the matrix.
        :type shape: int or tuple or list
        :param lower: Constant lower bound on the variable. May contain
            ``float("-inf")`` to denote unbounded elements.
        :param upper: Constant upper bound on the variable. May contain
            ``float("inf")`` to denote unbounded elements.
        """
        shape = load_shape(shape, squareMatrix=True)
        vec = SymmetricVectorization(shape)
        BaseVariable.__init__(self, name, vec, lower, upper)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Symmetric Variable"


class SkewSymmetricVariable(BaseVariable, AffineExpression):
    r"""A skew-symmetric matrix variable.

    Stored internally and passed to solvers as a skew-symmetric vectorization
    with only :math:`\frac{n(n-1)}{2}` entries.
    """

    def __init__(self, name, shape=(1, 1), lower=None, upper=None):
        """Create a :class:`SkewSymmetricVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of the matrix.
        :type shape: int or tuple or list
        :param lower: Constant lower bound on the variable. May contain
            ``float("-inf")`` to denote unbounded elements.
        :param upper: Constant upper bound on the variable. May contain
            ``float("inf")`` to denote unbounded elements.
        """
        shape = load_shape(shape, squareMatrix=True)
        vec = SkewSymmetricVectorization(shape)
        BaseVariable.__init__(self, name, vec, lower, upper)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Skew-symmetric Variable"


class HermitianVariable(BaseVariable, ComplexAffineExpression):
    r"""A hermitian matrix variable.

    Stored internally and passed to solvers as the horizontal concatenation of
    a real symmetric vectorization with :math:`\frac{n(n+1)}{2}` entries and a
    real skew-symmetric vectorization with :math:`\frac{n(n-1)}{2}` entries,
    resulting in a real vector with only :math:`n^2` entries total.
    """

    def __init__(self, name, shape):
        """Create a :class:`HermitianVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of the matrix.
        :type shape: int or tuple or list
        """
        shape = load_shape(shape, squareMatrix=True)
        vec = HermitianVectorization(shape)
        BaseVariable.__init__(self, name, vec)
        ComplexAffineExpression.__init__(
            self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Hermitian Variable"


class LowerTriangularVariable(BaseVariable, AffineExpression):
    r"""A lower triangular matrix variable.

    Stored internally and passed to solvers as a lower triangular vectorization
    with only :math:`\frac{n(n+1)}{2}` entries.
    """

    def __init__(self, name, shape=(1, 1), lower=None, upper=None):
        """Create a :class:`LowerTriangularVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of the matrix.
        :type shape: int or tuple or list
        :param lower: Constant lower bound on the variable. May contain
            ``float("-inf")`` to denote unbounded elements.
        :param upper: Constant upper bound on the variable. May contain
            ``float("inf")`` to denote unbounded elements.
        """
        shape = load_shape(shape, squareMatrix=True)
        vec = LowerTriangularVectorization(shape)
        BaseVariable.__init__(self, name, vec)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Lower Triangular Variable"


class UpperTriangularVariable(BaseVariable, AffineExpression):
    r"""An upper triangular matrix variable.

    Stored internally and passed to solvers as an upper triangular vectorization
    with only :math:`\frac{n(n+1)}{2}` entries.
    """

    def __init__(self, name, shape=(1, 1), lower=None, upper=None):
        """Create a :class:`UpperTriangularVariable`.

        :param str name: The variable's name, used for both string description
            and identification.
        :param shape: The shape of the matrix.
        :type shape: int or tuple or list
        :param lower: Constant lower bound on the variable. May contain
            ``float("-inf")`` to denote unbounded elements.
        :param upper: Constant upper bound on the variable. May contain
            ``float("inf")`` to denote unbounded elements.
        """
        shape = load_shape(shape, squareMatrix=True)
        vec = UpperTriangularVectorization(shape)
        BaseVariable.__init__(self, name, vec)
        AffineExpression.__init__(self, self.name, shape, {self: vec.identity})

    @classmethod
    def _get_type_string_base(cls):
        return "Lower Triangular Variable"


CONTINUOUS_VARTYPES = (RealVariable, ComplexVariable, SymmetricVariable,
                       SkewSymmetricVariable, HermitianVariable,
                       LowerTriangularVariable, UpperTriangularVariable)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019-2020 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Domain-specific container types."""

from collections import OrderedDict
from itertools import chain

from .apidoc import api_end, api_start
from .caching import cached_property
from .compat import MappingProxyType, MutableSet

_API_START = api_start(globals())
# -------------------------------


class OrderedSet(MutableSet):
    """A set that remembers its insertion order.

    >>> from picos.containers import OrderedSet as oset
    >>> o = oset([4, 3, 2, 1]); o
    OrderedSet([4, 3, 2, 1])
    >>> 3 in o
    True
    >>> o.update([5, 4, 3]); o
    OrderedSet([4, 3, 2, 1, 5])
    >>> list(o)
    [4, 3, 2, 1, 5]
    """

    def __init__(self, iterable=()):
        """Intialize the ordered set.

        :param iterable:
            Iterable to take initial elements from.
        """
        self._dict = OrderedDict((element, None) for element in iterable)

    # --------------------------------------------------------------------------
    # Special methods not implemented by MutableSet.
    # --------------------------------------------------------------------------

    def __str__(self):
        return "{{{}}}".format(", ".join(str(element) for element in self))

    def __repr__(self):
        return "OrderedSet([{}])".format(
            ", ".join(str(element) for element in self))

    # --------------------------------------------------------------------------
    # Abstract method implementations.
    # --------------------------------------------------------------------------

    def __contains__(self, key):
        return key in self._dict

    def __iter__(self):
        return iter(self._dict.keys())

    def __len__(self):
        return len(self._dict)

    def add(self, element):
        """Add an element to the set."""
        self._dict[element] = None

    def discard(self, element):
        """Discard an element from the set.

        If the element is not contained, do nothing.
        """
        self.pop(element, None)

    # --------------------------------------------------------------------------
    # Overridingns to improve performance over MutableSet's implementation.
    # --------------------------------------------------------------------------

    def clear(self):
        """Clear the set."""
        self._dict.clear()

    # --------------------------------------------------------------------------
    # Methods provided by set but not by MutableSet.
    # --------------------------------------------------------------------------

    def update(self, *iterables):
        """Update the set with elements from a number of iterables."""
        for iterable in iterables:
            for element in iterable:
                self._dict[element] = None

    difference = property(
        lambda self: self.__sub__,
        doc=set.difference.__doc__)

    difference_update = property(
        lambda self: self.__isub__,
        doc=set.difference_update.__doc__)

    intersection = property(
        lambda self: self.__and__,
        doc=set.intersection.__doc__)

    intersection_update = property(
        lambda self: self.__iand__,
        doc=set.intersection_update.__doc__)

    issubset = property(
        lambda self: self.__le__,
        doc=set.issubset.__doc__)

    issuperset = property(
        lambda self: self.__ge__,
        doc=set.issuperset.__doc__)

    symmetric_difference = property(
        lambda self: self.__xor__,
        doc=set.symmetric_difference.__doc__)

    symmetric_difference_update = property(
        lambda self: self.__ixor__,
        doc=set.symmetric_difference_update.__doc__)

    union = property(
        lambda self: self.__or__,
        doc=set.union.__doc__)


class frozendict(dict):
    """An immutable, hashable dictionary."""

    @classmethod
    def fromkeys(cls, iterable, value=None):
        """Overwrite :meth:`dict.fromkeys`."""
        return cls(dict.fromkeys(iterable, value))

    def __hash__(self):
        if not hasattr(self, "_hash"):
            self._hash = hash(tuple(sorted(self.items())))

        return self._hash

    def __str__(self):
        return dict.__repr__(self)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self)

    @property
    def _modify(self, *args, **kwargs):
        raise AttributeError(
            "Cannot modify a {}.".format(self.__class__.__name__))

    __delitem__ = _modify
    __setitem__ = _modify
    clear = _modify
    pop = _modify
    popitem = _modify
    setdefault = _modify
    update = _modify

    def copy(self):
        """Since :class:`frozendict` are immutable, returns self."""
        return self


class DetailedType:
    """Container for a pair of Python class and subtype.

    A container for a pair of a python class type and some logical subtype data
    structure, together called a detailed type.

    A detailed type is used when the mathematical type of an object must be
    distinguished more precisely than at the level of the python classes used to
    represent such mathematical objects. For instance, a single python class
    would be used for a type of expressions of varying dimensionality and
    subtypes would be used to distinguish further based on dimension.

    Instances of this class are treated exceptionally when used as a label of
    a :class:`RecordTree`: They are expanded into the class and the subtype as
    two seperate labels, making it convenient to store detailed types in trees.
    """

    def __init__(self, theClass, subtype):
        """Construct a :class:`DetailedType`.

        :param type theClass: The Python class part of the detailed type.
        :param object subtype: Additional type information.
        """
        if not hasattr(subtype, "_asdict"):
            raise TypeError("The given subtype of {} is not a namedtuple "
                "instance.".format(subtype))

        self.clstype = theClass
        self.subtype = subtype

    def __iter__(self):
        yield self.clstype
        yield self.subtype

    def __hash__(self):
        return hash((self.clstype, self.subtype))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def equals(self, other):
        """Whether two detailed types are the same."""
        return hash(self) == hash(other)

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, str(self))

    def __str__(self):
        subtypeArgsStr = "|".join("{}={}".format(key, val)
            for key, val in self.subtype._asdict().items())

        return "{}[{}]".format(self.clstype.__name__, subtypeArgsStr)

    def __add__(self, other):
        if isinstance(other, tuple):
            return tuple(self) + other
        elif isinstance(other, list):
            return list(self) + other
        else:
            return NotImplemented

    def __radd__(self, other):
        if isinstance(other, tuple):
            return other + tuple(self)
        elif isinstance(other, list):
            return other + list(self)
        else:
            return NotImplemented


class RecordTreeToken:
    """Base class for special :class:`RecordTree` value tokens."""

    def __init__(self):
        """Raise a :exc:`TypeError` on instanciation."""
        raise TypeError("{} may not be initialized.".format(
            self.__class__.__name__))


class RecordTree(object):
    """Labeled tree for storing records.

    An immutable labeled tree with values at the leaf nodes, where labels and
    values are arbitrary hashable python objects.

    An n-tuple whose first (n-1) elements are labels and whose last element is a
    value is called a record. Thus, every path from the root node to a leaf node
    represents one record.

    :class:`DetailedType` labels are treated exceptionally: They are expanded
    into the class and the subtype as two seperate labels.
    """

    class _NodeDict(dict):
        pass

    class NONE(RecordTreeToken):
        """Special :class:`RecordTree` value: No subtrees.

        If inserted at some (inner) node of the tree, the whole subtree starting
        at that node is deleted. If that node's parent node has no other
        children, then the parent node is deleted as well. This process is
        repeated iteratively up to the root node, which is never deleted.

        This is the only value that may be inserted at an inner node.

        This value cannot itself be stored in the tree as its insertion is
        always read as a deletion.
        """

        pass

    class ALL(RecordTreeToken):
        """Special :class:`RecordTree` value: Any subtrees.

        A special value that behaves as an arbitrary subtree during subtree
        checks.
        """

        pass

    @classmethod
    def _flatten(cls, path):
        for index, thing in enumerate(path):
            if isinstance(thing, DetailedType):
                return cls._flatten(path[:index] + thing + path[index+1:])
        return path

    @classmethod
    def _freeze(cls, value):
        """Make a label or value hashable."""
        if isinstance(value, list):
            newValue = tuple(value)
        elif isinstance(value, set):
            newValue = frozenset(value)
        elif isinstance(value, dict):
            newValue = frozendict(value)
        else:
            newValue = value

        try:
            hash(newValue)
        except Exception as error:
            raise TypeError(
                "Failed to freeze {} to a hashable type: Hashing {} gives: {}"
                .format(value, newValue, error))

        return newValue

    @staticmethod
    def _keyval_iterator(recordsOrDict):
        if isinstance(recordsOrDict, dict):
            return recordsOrDict.items()
        else:
            return ((rec[:-1], rec[-1]) for rec in recordsOrDict)

    @staticmethod
    def _record_iterator(recordsOrDict):
        if isinstance(recordsOrDict, dict):
            return ((key + (val,)) for key, val in recordsOrDict.items())
        else:
            return recordsOrDict

    def __init__(self, recordsOrDict=(), addValues=False, freeze=True):
        """Construct a :class:`RecordTree`.

        :param recordsOrDict:
            Data stored in the tree.
        :type recordsOrDict:
            dict or list(tuple)

        :param addValues:
            Add the (numeric) values of records stored in the same place in the
            tree, instead of replacing the value. If this is exactly a list of
            path tuples (precise types are required), then add values only for
            records below any of these paths instead. In either case, resulting
            values of zero are not stored in the tree.
        :type addValues:
            bool or list(tuple)

        :param bool freeze:
            Whether to transform mutable labels and values into hashable ones.
        """
        self._tree = self._NodeDict()

        if isinstance(addValues, list):
            addValues = [self._flatten(path) for path in addValues]

        def _add_values_at(path):
            if isinstance(addValues, list):
                return any(path[:end] in addValues for end in range(len(path)))
            else:
                return bool(addValues)

        for path, value in self._keyval_iterator(recordsOrDict):
            path = self._flatten(path)
            node = self._tree

            if freeze:
                path = tuple(self._freeze(thing) for thing in path)
                value = self._freeze(value)

            if value is not self.NONE and _add_values_at(path):
                if value == 0:
                    # Do not add a value equal to zero.
                    continue
                elif path in self:
                    oldValue = self[path]

                    if isinstance(oldValue, RecordTree):
                        raise LookupError("Can't add value '{}' at '{}': Path "
                            "leads to an inner node.".format(value, path))

                    value = oldValue + value

                    # If the sum is zero, delete the record instead.
                    if value == 0:
                        value = self.NONE

            if value is self.NONE:
                # Handle deletion of a subtree.
                clearNodes  = [node]
                clearLabels = []
                for label in path:
                    if label in node:
                        node = node[label]
                        clearNodes.insert(0, node)
                        clearLabels.insert(0, label)
                    else:
                        clearNodes = None
                        break

                if not clearNodes:
                    continue

                clearNodes.pop(0)
                for childLabel, node in zip(clearLabels, clearNodes):
                    node.pop(childLabel)
                    if node:
                        break
            else:
                # Handle insertion of a leaf (may replace a subtree).
                for label in path[:-1]:
                    node.setdefault(label, self._NodeDict())
                    node = node[label]

                    if not isinstance(node, self._NodeDict):
                        raise LookupError("Can't set value '{}' at '{}': Path "
                            "already contains a leaf.".format(value, path))

                node[path[-1]] = value

        self._hash = hash(self.set)

    @classmethod
    def _traverse(cls, node):
        if not isinstance(node, cls._NodeDict):
            # Not a node but a value.
            yield (node,)
            return
        elif not node:
            # Empty tree.
            return

        for label, child in node.items():
            for labels in cls._traverse(child):
                yield (label,) + labels

    @property
    def records(self):
        """Return an iterator over tuples, each representing one record."""
        return self._traverse(self._tree)

    @property
    def items(self):
        """Return an iterator over path/value pairs representing records."""
        return ((path[:-1], path[-1]) for path in self.records)

    @property
    def paths(self):
        """Return an iterator over paths, each representing one record."""
        return (path[:-1] for path in self.records)

    @cached_property
    def dict(self):
        """Return the tree as a read-only, tuple-indexed dictionary view.

        Every key/value pair of the returned dictionary is a record.
        """
        return MappingProxyType({path[:-1]: path[-1] for path in self.records})

    @cached_property
    def set(self):
        """Return a frozen set of tuples, each representing one record."""
        return frozenset(self.records)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        # Python 2 needs this in addition to __eq__.
        return hash(self) != hash(other)

    def __bool__(self):
        return bool(self._tree)

    def __len__(self):
        # TODO: Compute this on initialization.
        return len(list(self.records))

    def __contains__(self, path):
        if not isinstance(path, tuple):
            raise TypeError("{} indices must be tuples.".format(
                self.__class__.__name__))

        node = self._tree
        for label in path:
            if not isinstance(node, self._NodeDict) or label not in node:
                return False
            node = node[label]

        return True

    def _get(self, path, errorOnBadPath):
        # try:
        #     iter(path)
        # except TypeError:
        #     path = (path,)
        if not isinstance(path, tuple) and not isinstance(path, list):
            path = (path,)

        node = self._tree
        for label in path:
            if not isinstance(node, self._NodeDict) or label not in node:
                if errorOnBadPath:
                    raise LookupError(str(path))
                else:
                    return RecordTree()
            node = node[label]

        if isinstance(node, self._NodeDict):
            return RecordTree(
                {path[:-1]: path[-1] for path in self._traverse(node)})
        else:
            return node

    def __getitem__(self, path):
        return self._get(path, True)

    def get(self, path):
        """Return an empty :class:`RecordTree` if the path does not exist."""
        return self._get(path, False)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.dict)

    def __str__(self):
        return str(self.dict)

    def __le__(self, other):
        """Perform entrywise lower-or-equal-than comparison.

        Each left hand side path must be present on the right hand side, and the
        associated left hand side value must compare lower-or-equal-than the
        right hand side value.
        """
        if type(self) != type(other):
            return NotImplemented

        for path, value in self.items:
            if path not in other:
                return False

            if not value <= other[path]:
                return False

        return True

    def __ge__(self, other):
        """Perform entrywise greater-or-equal-than comparison.

        Each left hand side path must be present on the right hand side, and the
        associated left hand side value must compare greater-or-equal-than the
        right hand side value.
        """
        if type(self) != type(other):
            return NotImplemented

        for path, value in self.items:
            if path not in other:
                return False

            if not value >= other[path]:
                return False

        return True

    def __lshift__(self, other):
        """Perform subtree comparison.

        Each left hand side path must be present on the right hand side. If the
        special :class:`ALL` type is found as a value in the right hand side
        tree, it is treated as "all possible subtrees". All other values are not
        considered.
        """
        if not isinstance(other, RecordTree):
            return NotImplemented

        for path in self.paths:
            lhsNode = self._tree
            rhsNode = other._tree

            for label in path:
                if rhsNode is self.ALL:
                    break

                if label not in rhsNode:
                    return False

                assert label in lhsNode

                lhsNode = lhsNode[label]
                rhsNode = rhsNode[label]

        return True

    @staticmethod
    def _str(thing):
        return thing.__name__ if hasattr(thing, "__name__") else str(thing)

    @property
    def text(self):
        """Return the full tree as a multiline string."""
        keys, vals = [], []
        for path in self._traverse(self._tree):
            keys.append("/".join(self._str(label) for label in path[:-1]))
            vals.append(self._str(path[-1]))
        if not keys:
            return "Empty {} instance.".format(self.__class__.__name__)
        keyLen = max(len(key) for key in keys)
        valLen = max(len(val) for val in vals)
        return "\n".join(sorted(
            "{{:{}}} = {{:{}}}".format(keyLen, valLen).format(key, val)
            for key, val in zip(keys, vals)))

    def copy(self):
        """Create a shallow copy; the tree is copied, the values are not."""
        return self.__class__(self.records)

    def updated(self, recordsOrDict, addValues=False):
        """Create a shallow copy with modified records.

        :Example:

        >>> from picos.modeling.footprint import RecordTree as T
        >>> t = T({(1, 1, 1): 3, (1, 1, 2): 4, (1, 2, 1): 5}); t
        RecordTree({(1, 1, 1): 3, (1, 1, 2): 4, (1, 2, 1): 5})
        >>> t.updated({(1, 1, 1): "a", (2, 2): "b"}) # Change or add a record.
        RecordTree({(1, 1, 1): 'a', (1, 1, 2): 4, (1, 2, 1): 5, (2, 2): 'b'})
        >>> t.updated({(1,1,1): T.NONE}) # Delete a single record.
        RecordTree({(1, 1, 2): 4, (1, 2, 1): 5})
        >>> t.updated({(1,1): T.NONE}) # Delete multiple records.
        RecordTree({(1, 2, 1): 5})
        >>> t.updated([(1, 1, 1, T.NONE), (1, 1, 1, 1, 6)]) # Delete, then add.
        RecordTree({(1, 1, 2): 4, (1, 1, 1, 1): 6, (1, 2, 1): 5})
        >>> try: # Not possible to implicitly turn a leaf into an inner node.
        ...     t.updated([(1, 1, 1, 1, 6)])
        ... except LookupError as error:
        ...     print(error)
        Can't set value '6' at '(1, 1, 1, 1)': Path already contains a leaf.
        """
        return self.__class__(
            chain(self.records, self._record_iterator(recordsOrDict)),
            addValues)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Rotated second order cone constraints."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint

_API_START = api_start(globals())
# -------------------------------


class RSOCConstraint(Constraint):
    """Rotated second order cone membership constraint."""

    def __init__(self, normedExpression, upperBoundFactor1,
            upperBoundFactor2=None, customString=None):
        """Construct a :class:`RSOCConstraint`.

        :param ~picos.expressions.AffineExpression normedExpression:
            Expression under the norm.
        :param ~picos.expressions.AffineExpression upperBoundFactor1:
            First of the two scalar factors that make the upper bound on the
            normed expression.
        :param ~picos.expressions.AffineExpression upperBoundFactor2:
            Second of the two scalar factors that make the upper bound on the
            normed expression.
        :param str customString:
            Optional string description.
        """
        from ..expressions import AffineExpression

        if upperBoundFactor2 is None:
            upperBoundFactor2 = AffineExpression.from_constant(1)

        assert isinstance(normedExpression,  AffineExpression)
        assert isinstance(upperBoundFactor1, AffineExpression)
        assert isinstance(upperBoundFactor2, AffineExpression)
        assert len(upperBoundFactor1) == 1
        assert len(upperBoundFactor2) == 1

        self.ne  = normedExpression
        self.ub1 = upperBoundFactor1
        self.ub2 = upperBoundFactor2

        super(RSOCConstraint, self).__init__(
            self._get_type_term(), customString, printSize=True)

    def _get_type_term(self):
        return "RSOC"

    Subtype = namedtuple("Subtype", ("argdim",))

    def _subtype(self):
        return self.Subtype(len(self.ne))

    @classmethod
    def _cost(cls, subtype):
        return subtype.argdim + 2

    def _expression_names(self):
        yield "ne"
        yield "ub1"
        yield "ub2"

    def _str(self):
        a = glyphs.le(glyphs.squared(glyphs.norm(self.ne.string)),
            glyphs.clever_mul(self.ub1.string, self.ub2.string))
        b = glyphs.ge(self.ub1.string, 0)
        return glyphs.and_(a, b)

    def _get_size(self):
        return (len(self.ne) + 2, 1)

    def _get_slack(self):
        return self.ub1.value * self.ub2.value - (abs(self.ne)**2).value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`KullbackLeiblerConstraint`."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class KullbackLeiblerConstraint(Constraint):
    """Upper bound on a Kullback-Leibler divergence.

    This is the upper bound on a negative or relative entropy, both represented
    by :class:`~picos.expressions.NegativeEntropy`.
    """

    class ExpConeConversion(ConstraintConversion):
        """Kullback-Leibler to exponential cone constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable
            from . import AffineConstraint, ExpConeConstraint

            n = subtype.argdim

            yield ("var", RealVariable.make_var_type(dim=n, bnd=0), 1)
            yield ("con", AffineConstraint.make_type(dim=1, eq=False), 1)
            yield ("con", ExpConeConstraint.make_type(), n)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import ExponentialCone
            from ..modeling import Problem

            x = con.numerator
            y = con.denominator
            n = con.divergence.n
            b = con.upperBound

            P = Problem()

            u = P.add_variable("__u", n)
            P.add_constraint((u | 1) >= -b)

            for i in range(n):
                P.add_constraint((y[i] // x[i] // u[i]) << ExponentialCone())

            return P

    def __init__(self, divergence, upperBound):
        """Construct a :class:`KullbackLeiblerConstraint`.

        :param ~picos.expressions.NegativeEntropy divergence:
            Constrained expression.
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the expression.
        """
        from ..expressions import AffineExpression, NegativeEntropy

        assert isinstance(divergence, NegativeEntropy)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        self.divergence = divergence
        self.upperBound = upperBound

        super(KullbackLeiblerConstraint, self).__init__(divergence._typeStr)

    @property
    def numerator(self):
        """The :math:`x` of the divergence."""
        return self.divergence.x

    @cached_property
    def denominator(self):
        """The :math:`y` of the divergence, or :math:`1`."""
        from ..expressions import AffineExpression

        if self.divergence.y is None:
            return AffineExpression.from_constant(1, self.divergence.x.shape)
        else:
            return self.divergence.y

    Subtype = namedtuple("Subtype", ("argdim",))

    def _subtype(self):
        return self.Subtype(len(self.numerator))

    @classmethod
    def _cost(cls, subtype):
        # NOTE: Twice the argument dimension due to the denominator.
        return 2*subtype.argdim + 1

    def _expression_names(self):
        yield "divergence"
        yield "upperBound"

    def _str(self):
        return glyphs.le(self.divergence.string, self.upperBound.string)

    def _get_slack(self):
        return self.upperBound.value - self.divergence.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`RotatedSecondOrderCone`."""

import operator
from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..constraints import RSOCConstraint
from .data import convert_operands
from .exp_affine import AffineExpression
from .expression import refine_operands, validate_prediction
from .set import Set

_API_START = api_start(globals())
# -------------------------------


class RotatedSecondOrderCone(Set):
    r"""The (narrowed or widened) rotated second order cone.

    .. _rotatedcone:

    For :math:`n \in \mathbb{Z}_{\geq 3}` and :math:`p \in \mathbb{R}_{> 0}`,
    represents the convex cone

    .. math::

        \mathcal{R}_{p}^n = \left\{
            x \in \mathbb{R}^n
        ~\middle|~
            p x_1 x_2 \geq \sum_{i = 3}^n x_i^2 \land x_1, x_2 \geq 0
        \right\}.

    For :math:`p = 2`, this is the standard rotated second order cone
    :math:`\mathcal{R}^n` obtained by rotating the
    :class:`second order cone <picos.SecondOrderCone>` :math:`\mathcal{Q}^n`
    by :math:`\frac{\pi}{4}` in the :math:`(x_1, x_2)` plane.

    The default instance of this class has :math:`p = 1`, which can be
    understood as a narrowed version of the standard cone. This is more
    convenient for defining the primal problem but it should be noted that
    :math:`\mathcal{R}_{1}^n` is not self-dual, so working with
    :math:`p = 2` may seem more natural when the dual problem is of interest.

    :Dual cone:

    The dual cone is

    .. math::

        \left(\mathcal{R}_{p}^n\right)^* = \left\{
            x \in \mathbb{R}^n
        ~\middle|~
            \frac{4}{p} x_1 x_2 \geq \sum_{i = 2}^n x_i^2 \land x_1, x_2 \geq 0
        \right\}=\mathcal{R}_{4/p}^n.

    The cone is thus self-dual for :math:`p = 2`.
    """

    def __init__(self, p=1):
        """Construct a rotated second order cone.

        :param float p:
            The positive factor :math:`p` in the definition.
        """
        try:
            p = float(p)
        except Exception as error:
            raise TypeError("Failed to load the parameter 'p' as a float: {}"
                .format(error))

        if p <= 0:
            raise ValueError("The parameter 'p' must be positive.")

        self._p = p

        typeStr = "Rotated Second Order Cone"
        if p < 2:
            typeStr = "Narrowed " + typeStr
        elif p > 2:
            typeStr = "Widened " + typeStr

        symbStr = glyphs.set(glyphs.sep(
            glyphs.col_vectorize("u", "v", "x"), glyphs.and_(
            glyphs.le(
                glyphs.squared(glyphs.norm("x")),
                glyphs.clever_mul(glyphs.scalar(p), glyphs.mul("u", "v"))),
            glyphs.ge("u", 0))))

        Set.__init__(self, typeStr, symbStr)

    @property
    def p(self):
        """A narrowing (:math:`p < 2`) or widening (:math:`p > 2`) factor."""
        return self._p

    def _get_variables(self):
        return set()

    def _replace_variables(self):
        return self

    Subtype = namedtuple("Subtype", ())

    def _get_subtype(self):
        return self.Subtype()

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        if relation == operator.__rshift__:
            if issubclass(other.clstype, AffineExpression):
                if other.subtype.dim >= 3:
                    return RSOCConstraint.make_type(other.subtype.dim - 2)

        return NotImplemented

    @convert_operands()
    @validate_prediction
    @refine_operands()
    def __rshift__(self, element):
        if isinstance(element, AffineExpression):
            if len(element) < 3:
                raise TypeError("Elements of the rotated second order cone must"
                    " be at least three-dimensional.")

            element = element.vec

            return RSOCConstraint(element[2:], self.p * element[0], element[1])
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C)      2018 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`VectorNormConstraint`."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class VectorNormConstraint(Constraint):
    """Bound on a (generalized) vector :math:`p`-norm."""

    class Conversion(ConstraintConversion):
        """Bound on a (generalized) :math:`p`-norm constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable as RV
            from . import AffineConstraint as AC, GeometricMeanConstraint as GM

            n, num, den, relation = subtype
            p = float(num) / float(den)

            assert p >= 0.0

            if relation == Constraint.LE:
                assert p >= 1.0

                if p == 1.0:
                    yield ("var", RV.make_var_type(dim=n, bnd=0), 1)
                    yield ("con", AC.make_type(dim=n, eq=False), 2)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
                elif p == float("inf"):
                    yield ("con", AC.make_type(dim=n, eq=False), 2)
                else:  # 1 < p < inf
                    yield ("var", RV.make_var_type(dim=n, bnd=0), 2)
                    yield ("con", AC.make_type(dim=1, eq=False), 2*n)
                    yield ("con", GM.make_type(argdim=num), n)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
            else:
                assert relation == Constraint.GE
                assert p <= 1.0

                if p == 1.0:
                    yield ("con", AC.make_type(dim=n, eq=False), 1)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
                elif p == float("-inf"):
                    yield ("con", AC.make_type(dim=n, eq=False), 1)
                elif p > 0:
                    yield ("var", RV.make_var_type(dim=n, bnd=0), 1)
                    yield ("con", GM.make_type(argdim=den), n)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)
                else:  # -inf < p < 0
                    yield ("var", RV.make_var_type(dim=n, bnd=0), 1)
                    yield ("con", GM.make_type(argdim=(num + den)), n)
                    yield ("con", AC.make_type(dim=1, eq=False), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import GeometricMean
            from ..modeling import Problem

            norm     = con.norm
            relation = con.relation
            rhs      = con.rhs

            p = norm.p
            m = len(norm.x)

            P = Problem()

            if relation == Constraint.LE:
                if p == 1:
                    v = P.add_variable('__v', m)
                    P.add_constraint(norm.x[:] <= v)
                    P.add_constraint(-norm.x[:] <= v)
                    P.add_constraint((1 | v) <= rhs)
                elif p == float('inf'):
                    P.add_constraint(norm.x <= rhs)
                    P.add_constraint(-norm.x <= rhs)
                else:
                    x = P.add_variable('__x', m)
                    v = P.add_variable('__v', m)
                    amb = norm.pnum - norm.pden
                    b = norm.pden
                    oneamb = '|1|(' + str(amb) + ',1)'
                    oneb = '|1|(' + str(b) + ',1)'
                    for i in range(m):
                        if amb > 0:
                            if b == 1:
                                vec = (v[i]) // (rhs * oneamb)
                            else:
                                vec = (v[i] * oneb) // (rhs * oneamb)
                        else:
                            # TODO: This shouldn't be possible?
                            if b == 1:
                                vec = v[i]
                            else:
                                vec = (v[i] * oneb)
                        P.add_constraint(norm.x[i] <= x[i])
                        P.add_constraint(-norm.x[i] <= x[i])
                        P.add_constraint(x[i] <= GeometricMean(vec))
                    P.add_constraint((1 | v) <= rhs)
            else:
                if p == 1:
                    P.add_constraint(norm.x >= 0)
                    P.add_constraint((1 | norm.x) >= rhs)
                elif p == float('-inf'):
                    P.add_constraint(norm.x >= rhs)
                elif p >= 0:
                    v = P.add_variable('__v', m)
                    bma = -(norm.pnum - norm.pden)
                    a = norm.pnum
                    onebma = '|1|(' + str(bma) + ',1)'
                    onea = '|1|(' + str(a) + ',1)'
                    for i in range(m):
                        if a == 1:
                            vec = (norm.x[i]) // (rhs * onebma)
                        else:
                            vec = (norm.x[i] * onea) // (rhs * onebma)
                        P.add_constraint(v[i] <= GeometricMean(vec))
                    P.add_constraint(rhs <= (1 | v))
                else:
                    v = P.add_variable('__v', m)
                    b = abs(norm.pden)
                    a = abs(norm.pnum)
                    oneb = '|1|(' + str(b) + ',1)'
                    onea = '|1|(' + str(a) + ',1)'
                    for i in range(m):
                        if a == 1 and b == 1:
                            vec = (norm.x[i]) // (v[i])
                        elif a > 1 and b == 1:
                            vec = (norm.x[i] * onea) // (v[i])
                        elif a == 1 and b > 1:
                            vec = (norm.x[i]) // (v[i] * oneb)
                        else:
                            vec = (norm.x[i] * onea) // (v[i] * oneb)
                        P.add_constraint(rhs <= GeometricMean(vec))
                    P.add_constraint((1 | v) <= rhs)

            return P

    def __init__(self, norm, relation, rhs):
        """Construct a :class:`VectorNormConstraint`.

        :param ~picos.expressions.Norm norm:
            Left hand side expression.
        :param str relation:
            Constraint relation symbol.
        :param ~picos.expressions.AffineExpression rhs:
            Right hand side expression.
        """
        from ..expressions import AffineExpression, Norm

        assert isinstance(norm, Norm)
        assert isinstance(rhs, AffineExpression)
        assert relation in self.LE + self.GE
        assert len(rhs) == 1

        assert norm.q == norm.p, \
            "Can't create a p-norm constraint for a (p,q)-norm."

        if relation == self.LE:
            assert norm.convex, \
                "Upper bounding a p-norm requires p s.t. the norm is convex."

        if relation == self.GE:
            assert norm.concave, \
                "Lower bounding a p-norm requires p s.t. the norm is concave."

        self.norm     = norm
        self.relation = relation
        self.rhs      = rhs

        super(VectorNormConstraint, self).__init__(norm._typeStr)

    Subtype = namedtuple("Subtype", ("argdim", "num", "den", "relation"))

    def _subtype(self):
        return self.Subtype(
            len(self.norm.x), self.norm.pnum, self.norm.pden, self.relation)

    @classmethod
    def _cost(cls, subtype):
        return subtype.argdim + 1

    def _expression_names(self):
        yield "norm"
        yield "rhs"

    def _str(self):
        if self.relation == self.LE:
            str = glyphs.le(self.norm.string, self.rhs.string)
        else:
            str = glyphs.ge(self.norm.string, self.rhs.string)

        if self.norm.p < 1:
            return glyphs.and_(str, glyphs.ge(self.norm.x.string, 0))
        else:
            return str

    def _get_slack(self):
        if self.relation == self.LE:
            return self.rhs.value - self.norm.value
        else:
            return self.norm.value - self.rhs.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Affine constraint types."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_property
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class AffineConstraint(Constraint):
    """An equality or inequality between two affine expressions."""

    def __init__(self, lhs, relation, rhs, customString=None):
        """Construct an :class:`AffineConstraint`.

        :param ~picos.expressions.AffineExpression lhs:
            Left hand side expression.
        :param str relation:
            Constraint relation symbol.
        :param ~picos.expressions.AffineExpression rhs:
            Right hand side expression.
        :param str customString:
            Optional string description.
        """
        from ..expressions import AffineExpression

        assert isinstance(lhs, AffineExpression)
        assert isinstance(rhs, AffineExpression)
        assert relation in self.LE + self.GE + self.EQ
        assert lhs.size == rhs.size

        self.lhs      = lhs
        self.rhs      = rhs
        self.relation = relation

        super(AffineConstraint, self).__init__(
            "Affine", customString, printSize=True)

    Subtype = namedtuple("Subtype", ("dim", "eq"))

    def _subtype(self):
        return self.Subtype(len(self.lhs), self.relation == self.EQ)

    @property
    def smaller(self):
        """Smaller-or-equal side of the constraint.

        The smaller-or-equal side expression in case of an inequality, otherwise
        the left hand side.
        """
        return self.rhs if self.relation == self.GE else self.lhs

    @property
    def greater(self):
        """Greater-or-equal side of the constraint.

        The greater-or-equal side expression in case of an inequality, otherwise
        the right hand side.
        """
        return self.lhs if self.relation == self.GE else self.rhs

    @cached_property
    def le0(self):
        """Expression constrained to be lower than or equal to zero.

        The expression posed to be less than or equal to zero in case of an
        inequality, otherwise the left hand side minus the right hand side.
        """
        if self.relation == self.GE:
            return self.rhs - self.lhs
        else:
            return self.lhs - self.rhs

    @cached_property
    def ge0(self):
        """Expression constrained to be greater than or equal to zero.

        The expression posed to be greater than or equal to zero in case of an
        inequality, otherwise the right hand side minus the left hand side.
        """
        if self.relation == self.GE:
            return self.lhs - self.rhs
        else:
            return self.rhs - self.lhs

    @classmethod
    def _cost(cls, subtype):
        return subtype.dim

    def _expression_names(self):
        yield "lhs"
        yield "rhs"

    def _str(self):
        if self.relation == self.LE:
            return glyphs.le(self.lhs.string, self.rhs.string)
        elif self.relation == self.GE:
            return glyphs.ge(self.lhs.string, self.rhs.string)
        else:
            return glyphs.eq(self.lhs.string, self.rhs.string)

    def _get_size(self):
        return self.lhs.size

    def _get_slack(self):
        if self.relation == self.LE:
            delta = self.rhs.value - self.lhs.value
        else:
            delta = self.lhs.value - self.rhs.value

        return -abs(delta) if self.relation == self.EQ else delta

    def bounded_linear_form(self):
        """Bounded linear form of the constraint.

        Separates the constraint into a linear function on the left hand side
        and a constant bound on the right hand side.

        :returns: A pair ``(linear, bound)`` where ``linear`` is a pure linear
            expression and ``bound`` is a constant expression.
        """
        linear = self.lhs - self.rhs
        bound  = -linear.cst
        linear = linear + bound

        return (linear, bound)

    def sparse_Ab_rows(self, varOffsetMap, indexFunction=None):
        """Sparse representation of the constraint's bounded linear form.

        A sparse list representation of the constraint, given a mapping of PICOS
        variables to column offsets (or alternatively given an index function).

        The constraint is brought into a bounded linear form A • b, where • is
        one of ≤, ≥, or =, depending on the constraint relation, and the rows
        returned correspond to the matrix [A|b].

        :param dict varOffsetMap: Maps variables or variable start indices to
            column offsets.
        :param indexFunction: Instead of adding the local variable index to the
            value returned by varOffsetMap, use the return value of this
            function, that takes as argument the variable and its local index,
            as the "column index", which need not be an integer. When this
            parameter is passed, the parameter varOffsetMap is ignored.
        :returns: A list of triples (J, V, c) where J contains column indices
            (representing scalar variables), V contains coefficients for each
            column index and c is a constant term. Each entry of the list
            represents a row in a constraint matrix.
        """
        lhs  = self.lhs - self.rhs
        rows = lhs.sparse_rows(varOffsetMap, indexFunction=indexFunction)

        for localConIndex in range(len(lhs)):
            rows[localConIndex][2] = -rows[localConIndex][2]

        return rows


class ComplexAffineConstraint(Constraint):
    """An equality between affine expressions, at least one being complex."""

    class RealConversion(ConstraintConversion):
        """Complex affine equality to real affine equality conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            yield ("con",
                AffineConstraint.make_type(dim=2*subtype.dim, eq=True), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem

            P = Problem()
            P.add_constraint((con.lhs.real // con.lhs.imag)
                == (con.rhs.real // con.rhs.imag))

            return P

        @classmethod
        def dual(cls, auxVarPrimals, auxConDuals, options):
            """Implement :meth:`~.constraint.ConstraintConversion.dual`."""
            assert len(auxConDuals) == 1

            auxConDual = auxConDuals[0]
            if auxConDual is None:
                return None
            else:
                n = auxConDual.size[0] // 2
                return auxConDual[:n, :] + 1j*auxConDual[n:, :]

    def __init__(self, lhs, rhs, customString=None):
        """Construct a :class:`ComplexAffineConstraint`.

        :param ~picos.expressions.AffineExpression lhs:
            Left hand side expression.
        :param ~picos.expressions.AffineExpression rhs:
            Right hand side expression.
        :param str customString:
            Optional string description.
        """
        from ..expressions import ComplexAffineExpression

        assert isinstance(lhs, ComplexAffineExpression)
        assert isinstance(rhs, ComplexAffineExpression)
        assert lhs.size == rhs.size

        self.lhs = lhs
        self.rhs = rhs

        super(ComplexAffineConstraint, self).__init__(
            "Complex Equality", customString, printSize=True)

    Subtype = namedtuple("Subtype", ("dim",))

    def _subtype(self):
        return self.Subtype(len(self.lhs))

    @classmethod
    def _cost(cls, subtype):
        return 2*subtype.dim

    def _expression_names(self):
        yield "lhs"
        yield "rhs"

    def _str(self):
        return glyphs.eq(self.lhs.string, self.rhs.string)

    def _get_size(self):
        return self.lhs.size

    def _get_slack(self):
        return -abs(self.lhs.value - self.rhs.value)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2017 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`SMCPSolver`."""

# ------------------------------------------------------------------------------
# This file is a skeleton implementation for the SMCP solver.
# Currently, the funcional implementation is found in CVXOPTSolver.
# ------------------------------------------------------------------------------

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import AffineConstraint, LMIConstraint
from ..expressions import CONTINUOUS_VARTYPES, AffineExpression
from ..modeling.footprint import Specification
from .solver import Solver
from .solver_cvxopt import CVXOPTSolver

_API_START = api_start(globals())
# -------------------------------


class SMCPSolver(CVXOPTSolver):
    """Interface to the SMCP solver.

    Most of the logic is implemented in the
    :class:`CVXOPTSolver <picos.solvers.solver_cvxopt.CVXOPTSolver>` base class.
    """

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression],
        variables=CONTINUOUS_VARTYPES,
        constraints=[
            AffineConstraint,
            # TODO: See below.
            # SOCConstraint,
            # RSOCConstraint,
            LMIConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        # TODO: SMCP formally supports problems that are not SDPs, but it is
        #       known to perform poorly for them, sometimes even returning
        #       incorrect solutions. For now, only support SDPs.
        if not ("con", LMIConstraint) in footprint:
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 2.0  # Experimental free/open source solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("smcp")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "smcp", "SMCP", "Sparse Matrix Cone Program Solver"

    @property
    def is_smcp(self):
        """Whether to implement SMCP instead of CVXOPT."""
        return True


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`AbsoluteValueConstraint`."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class AbsoluteValueConstraint(Constraint):
    """Upper bound on an absolute value."""

    # TODO: Improve performance: Add two scalar constraints.
    class AffineConversion(ConstraintConversion):
        """Upper bound on an absolute value to affine inequality conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from . import AffineConstraint

            yield ("con", AffineConstraint.make_type(dim=2, eq=False), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..modeling import Problem

            P = Problem()
            P.add_constraint((con.signedScalar // -con.signedScalar)
                <= (con.upperBound // con.upperBound))

            return P

        @classmethod
        def dual(cls, auxVarPrimals, auxConDuals, options):
            """Implement :meth:`~.constraint.ConstraintConversion.dual`."""
            assert len(auxConDuals) == 1

            if auxConDuals[0] is None:
                return None
            else:
                return auxConDuals[0][1] - auxConDuals[0][0]

    def __init__(self, signedScalar, upperBound):
        """Construct an :class:`AbsoluteValueConstraint`.

        :param ~picos.expressions.AffineExpression signedScalar:
            A scalar expression.
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the expression.
        """
        from ..expressions import AffineExpression

        assert isinstance(signedScalar, AffineExpression)
        assert isinstance(upperBound, AffineExpression)
        assert len(signedScalar) == 1
        assert len(upperBound) == 1

        self.signedScalar = signedScalar
        self.upperBound   = upperBound

        super(AbsoluteValueConstraint, self).__init__("Absolute Value")

    Subtype = namedtuple("Subtype", ())

    def _subtype(self):
        return self.Subtype()

    @classmethod
    def _cost(cls, subtype):
        return 2

    def _expression_names(self):
        yield "signedScalar"
        yield "upperBound"

    def _str(self):
        return glyphs.le(
            glyphs.abs(self.signedScalar.string), self.upperBound.string)

    def _get_size(self):
        return (1, 1)

    def _get_slack(self):
        return self.upperBound.value - abs(self.signedScalar.value)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`SCIPSolver`."""

import cvxopt

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import (AffineConstraint, ConvexQuadraticConstraint,
                           NonconvexQuadraticConstraint, RSOCConstraint,
                           SOCConstraint)
from ..expressions import AffineExpression, BinaryVariable, IntegerVariable
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_INFEASIBLE, PS_UNBOUNDED,
                                 PS_UNKNOWN, SS_INFEASIBLE, SS_OPTIMAL,
                                 SS_PREMATURE, SS_UNKNOWN)
from .solver import Solver

_API_START = api_start(globals())
# -------------------------------


class SCIPSolver(Solver):
    """Interface to the SCIP solver via the PySCIPOpt Python interface."""

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression],
        variables=Specification.ALL,
        constraints=[
            AffineConstraint,
            SOCConstraint,
            RSOCConstraint,
            ConvexQuadraticConstraint,
            NonconvexQuadraticConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        # HACK: SCIP does not produce dual values (except for LPs, where they
        #       are known to be sometimes incorrect). For now pretend SCIP does
        #       not support continuous problems at all (unless there are
        #       nonconvex quadratics where no other solver works).
        # TODO: Support all continuous problems when ``duals=False``.
        if footprint.continuous \
        and not ("con", NonconvexQuadraticConstraint) in footprint:
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 1.0  # Stable free/open source solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("pyscipopt")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "scip", "SCIP", "SCIP Optimization Suite"

    def __init__(self, problem):
        """Initialize a SCIP solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        super(SCIPSolver, self).__init__(problem)

        self._scipVar = dict()
        """Maps PICOS variable indices to SCIP variables."""

        self._scipCons = dict()
        """
        Maps PICOS constraints to lists of SCIP constraints.

        For PICOS second order cone constraints, the first entry in the list is
        a SCIP quadratic constraint and the second entry is a SCIP linear
        auxiliary constraint.
        """

        self._scipQuadObjAuxVar = None
        """A SCIP auxiliary variable to support a PICOS quadratic objective."""

        self._scipQuadObjAuxCon = None
        """A SCIP auxiliary constraint to support a PICOS quadr. objective."""

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        self.int = None

        self._scipVar.clear()
        self._scipCons.clear()

        self._scipQuadObjAuxVar = None
        self._scipQuadObjAuxCon = None

    def _make_scip_var_names(self, picosVar, localIndex=None):
        """Make SCIP variable names.

        Converts a PICOS variable to a list of SCIP variable names, each
        corresponding to one scalar variable contained in the PICOS variable.
        If localIndex is given, then only the name of the SCIP variable
        representing the scalar variable with that offset is returned.
        The name format is "picosName[localIndex]".
        """
        # TODO: This function appears in multiple solvers, move it to the Solver
        #       base class as "_make_scalar_var_names".
        if localIndex is not None:
            return "{}[{}]".format(picosVar.name, localIndex)
        else:
            return [
                self._make_scip_var_names(picosVar, localIndex)
                for localIndex in range(len(picosVar))]

    def _import_variable(self, picosVar):
        from math import ceil, floor

        dim = picosVar.dim
        inf = self.int.infinity()

        # Retrieve type code.
        if isinstance(picosVar, (BinaryVariable, IntegerVariable)):
            scipVarType = "I"
        else:
            scipVarType = "C"

        # Retrieve bounds.
        lowerBounds = [-inf]*dim
        upperBounds = [inf]*dim
        lower, upper = picosVar.bound_dicts
        for i, b in lower.items():
            lowerBounds[i] = b
        for i, b in upper.items():
            upperBounds[i] = b

        # Refine bounds.
        if isinstance(picosVar, IntegerVariable):
            lowerBounds = [int(ceil(b)) for b in lowerBounds]
            upperBounds = [int(floor(b)) for b in upperBounds]

        # Give names.
        scipNames = self._make_scip_var_names(picosVar)

        # Import scalar variables.
        for i in range(dim):
            self._scipVar[picosVar.id_at(i)] = self.int.addVar(
                name=scipNames[i], vtype=scipVarType,
                lb=lowerBounds[i], ub=upperBounds[i])

    def _import_variable_values(self):
        # TODO: Import variable values with SCIP.
        raise NotImplementedError

    def _reset_variable_values(self):
        # TODO: Import variable values with SCIP.
        raise NotImplementedError

    def _affinexp_pic2scip(self, picosExpression):
        """Transform an affine expression from PICOS to SCIP.

        Multidimensional PICOS expressions are converted to a number of scalar
        SCIP expressions.

        :returns: A :class:`list` of :class:`SCIP expressions
            <pyscipopt.scip.Expr>`.
        """
        from pyscipopt.scip import Expr

        if picosExpression is None:
            yield Expr()
            return

        for scipVars, coefficients, constant in picosExpression.sparse_rows(
                None, indexFunction=lambda picosVar, localIndex:
                self._scipVar[picosVar.id_at(localIndex)]):
            scipExpression = Expr()
            for scipVar, coefficient in zip(scipVars, coefficients):
                scipExpression += coefficient * scipVar
            scipExpression += constant
            yield scipExpression

    def _scalar_affinexp_pic2scip(self, picosExpression):
        """Transform a PICOS scalar affine expression into a SCIP expression.

        :returns: A :class:`SCIP expression <pyscipopt.scip.Expr>`.
        """
        assert len(picosExpression) == 1
        return next(self._affinexp_pic2scip(picosExpression))

    def _quadexp_pic2scip(self, picosExpression):
        # Convert affine part.
        scipExpression = self._scalar_affinexp_pic2scip(picosExpression.aff)

        # Convert quadratic part.
        for (pVar1, pVar2), pCoefficients \
        in picosExpression.quadratic_forms.items():
            for sparseIndex in range(len(pCoefficients)):
                localVar1Index   = pCoefficients.I[sparseIndex]
                localVar2Index   = pCoefficients.J[sparseIndex]
                localCoefficient = pCoefficients.V[sparseIndex]
                scipVar1 = self._scipVar[pVar1.id_at(localVar1Index)]
                scipVar2 = self._scipVar[pVar2.id_at(localVar2Index)]
                scipExpression += localCoefficient * scipVar1 * scipVar2

        return scipExpression

    def _import_linear_constraint(self, picosConstraint):
        assert isinstance(picosConstraint, AffineConstraint)

        scipCons = []
        picosLHS = picosConstraint.lhs - picosConstraint.rhs
        for scipLHS in self._affinexp_pic2scip(picosLHS):
            if picosConstraint.is_increasing():
                scipCons.append(self.int.addCons(scipLHS <= 0.0))
            elif picosConstraint.is_decreasing():
                scipCons.append(self.int.addCons(scipLHS >= 0.0))
            elif picosConstraint.is_equality():
                scipCons.append(self.int.addCons(scipLHS == 0.0))
            else:
                assert False, "Unexpected constraint relation."

        return scipCons

    def _import_quad_constraint(self, picosConstraint):
        assert isinstance(picosConstraint,
            (ConvexQuadraticConstraint, NonconvexQuadraticConstraint))

        scipLHS  = self._quadexp_pic2scip(picosConstraint.le0)
        scipCons = [self.int.addCons(scipLHS <= 0.0)]

        return scipCons

    def _import_socone_constraint(self, picosConstraint):
        scipCons = []
        scipQuadLHS = self._quadexp_pic2scip(
            (picosConstraint.ne | picosConstraint.ne)
            - (picosConstraint.ub * picosConstraint.ub))
        scipCons.append(self.int.addCons(scipQuadLHS <= 0.0))

        scipAuxLHS = self._scalar_affinexp_pic2scip(picosConstraint.ub)
        if scipAuxLHS.degree() > 0:
            scipCons.append(self.int.addCons(scipAuxLHS >= 0.0))

        return scipCons

    def _import_rscone_constraint(self, picosConstraint):
        scipCons = []
        scipLHS = self._quadexp_pic2scip(
            (picosConstraint.ne | picosConstraint.ne)
            - (picosConstraint.ub1 * picosConstraint.ub2))
        scipCons.append(self.int.addCons(scipLHS <= 0.0))

        # Make sure that either the RHS is constant, or one of the two
        # expressions is non-negative.
        scipAuxLHS = self._scalar_affinexp_pic2scip(picosConstraint.ub1)
        if scipAuxLHS.degree() > 0:
            scipCons.append(self.int.addCons(scipAuxLHS >= 0.0))
        else:
            scipAuxLHS = self._scalar_affinexp_pic2scip(picosConstraint.ub2)
            if scipAuxLHS.degree() > 0:
                scipCons.append(self.int.addCons(scipAuxLHS >= 0.0))

        return scipCons

    def _import_constraint(self, picosConstraint):
        # Import constraint based on type.
        if isinstance(picosConstraint, AffineConstraint):
            scipCons = self._import_linear_constraint(picosConstraint)
        elif isinstance(picosConstraint,
                (ConvexQuadraticConstraint, NonconvexQuadraticConstraint)):
            scipCons = self._import_quad_constraint(picosConstraint)
        elif isinstance(picosConstraint, SOCConstraint):
            scipCons = self._import_socone_constraint(picosConstraint)
        elif isinstance(picosConstraint, RSOCConstraint):
            scipCons = self._import_rscone_constraint(picosConstraint)
        else:
            assert False, "Constraint type belongs to unsupported problem type."

        # Map PICOS constraints to lists of SCIP constraints.
        self._scipCons[picosConstraint] = scipCons

    def _import_objective(self):
        picosSense, picosObjective = self.ext.no

        # Retrieve objective sense.
        if picosSense == "min":
            scipSense = "minimize"
        else:
            assert picosSense == "max"
            scipSense = "maximize"

        # Import objective function.
        scipObjective = self._scalar_affinexp_pic2scip(picosObjective)

        # HACK: Remove a constant term from the objective as this is not allowed
        #       by SCIP under Python 2. (Not sure if the term is stored or
        #       dropped under Python 3, but no exception is thrown there.)
        from pyscipopt.scip import Term
        constantTerm = Term()
        if constantTerm in scipObjective.terms:
            scipObjective.terms.pop(constantTerm)

        self.int.setObjective(scipObjective, scipSense)

    def _import_problem(self):
        import pyscipopt as scip

        # Create a problem instance.
        self.int = scip.Model()

        # Import variables.
        for variable in self.ext.variables.values():
            self._import_variable(variable)

        # Import constraints.
        for constraint in self.ext.constraints.values():
            self._import_constraint(constraint)

        # Set objective.
        self._import_objective()

    def _update_problem(self):
        # TODO: Support all problem updates supported by SCIP.
        raise NotImplementedError

    def _solve(self):
        import pyscipopt as scip

        # TODO: Give Problem an interface for checks like this.
        isLP = isinstance(self.ext.no.function, AffineExpression) \
            and all(isinstance(constraint, AffineConstraint)
            for constraint in self.ext.constraints.values())

        # Reset options.
        self.int.resetParams()

        # verbosity
        picosVerbosity = self.verbosity()
        if picosVerbosity <= -1:
            scipVerbosity = 0
        elif picosVerbosity == 0:
            scipVerbosity = 2
        elif picosVerbosity == 1:
            scipVerbosity = 3
        elif picosVerbosity >= 2:
            scipVerbosity = 5
        self.int.setIntParam("display/verblevel", scipVerbosity)

        # TODO: "numerics/lpfeastol" has been replaced in SCIP 7.0.0 (PySCIPOpt
        #       3.0.0) by "numerics/lpfeastolfactor", which is multiplied with
        #       some unknown default feasibility (maybe "numerics/feastol"?).
        #       Looking at it now it is not clear to me which of SCIP's
        #       feasibility tolerances does what exactly and whether they are
        #       absolute or relative measures. The workaround is to maintain
        #       PICOS' old guess but ignore abs_prim_fsb_tol with the new SCIP.
        if int(scip.__version__.split(".")[0]) < 3:
            # abs_prim_fsb_tol
            if self.ext.options.abs_prim_fsb_tol is not None:
                self.int.setRealParam("numerics/lpfeastol",
                    self.ext.options.abs_prim_fsb_tol)

        # abs_dual_fsb_tol
        if self.ext.options.abs_dual_fsb_tol is not None:
            self.int.setRealParam("numerics/dualfeastol",
                self.ext.options.abs_dual_fsb_tol)

        # rel_prim_fsb_tol, rel_dual_fsb_tol
        relFsbTols = [tol for tol in (self.ext.options.rel_prim_fsb_tol,
            self.ext.options.rel_dual_fsb_tol) if tol is not None]
        if relFsbTols:
            self.int.setRealParam("numerics/feastol", min(relFsbTols))

        # abs_ipm_opt_tol, abs_bnb_opt_tol
        absOptTols = [tol for tol in (self.ext.options.abs_ipm_opt_tol,
            self.ext.options.abs_bnb_opt_tol) if tol is not None]
        if absOptTols:
            self.int.setRealParam("limits/absgap", min(absOptTols))

        # rel_ipm_opt_tol, rel_bnb_opt_tol
        relOptTols = [tol for tol in (self.ext.options.rel_ipm_opt_tol,
            self.ext.options.rel_bnb_opt_tol) if tol is not None]
        if relOptTols:
            self.int.setRealParam("limits/gap", min(relOptTols))

        # timelimit
        if self.ext.options.timelimit is not None:
            self.int.setRealParam("limits/time", self.ext.options.timelimit)

        # treememory
        if self.ext.options.treememory is not None:
            self.int.setRealParam("limits/memory", self.ext.options.treememory)

        # max_fsb_nodes
        if self.ext.options.max_fsb_nodes is not None:
            self.int.setRealParam(
                "limits/solutions", float(self.ext.options.max_fsb_nodes))

        # Handle SCIP-specific options.
        for option, value in self.ext.options.scip_params.items():
            try:
                if isinstance(value, bool):
                    self.int.setBoolParam(option, value)
                elif isinstance(value, str):
                    if len(value) == 1:
                        try:
                            self.int.setCharParam(option, ord(value))
                        except LookupError:
                            self.int.setStringParam(option, value)
                    else:
                        self.int.setStringParam(option, value)
                elif isinstance(value, float):
                    self.int.setRealParam(option, value)
                elif isinstance(value, int):
                    try:
                        self.int.setIntParam(option, value)
                    except LookupError:
                        try:
                            self.int.setLongintParam(option, value)
                        except LookupError:
                            self.int.setRealParam(option, float(value))
            except KeyError:
                self._handle_bad_option_value("scip_params",
                    "SCIP option '{}' does not exist.".format(option))
            except ValueError:
                self._handle_bad_option_value("scip_params",
                    "Invalid value '{}' for SCIP option '{}'."
                    .format(value, option))
            except LookupError:
                self._handle_bad_option_value("scip_params",
                    "Failed to guess type of SCIP option '{}'.".format(option))

        # Handle unsupported options.
        self._handle_unsupported_options(
            "hotstart", "lp_node_method", "lp_root_method", "max_iterations")

        # In the case of a pure LP, disable presolve to get duals.
        if self.ext.options.duals is not False and isLP:
            self._debug("Disabling SCIP's presolve, heuristics, and propagation"
                " features to allow for LP duals.")

            self.int.setPresolve(scip.SCIP_PARAMSETTING.OFF)
            self.int.setHeuristics(scip.SCIP_PARAMSETTING.OFF)

            # Note that this is a helper to set options, so they get reset at
            # the beginning of the function instead of in the else-scope below.
            self.int.disablePropagation()
        else:
            self.int.setPresolve(scip.SCIP_PARAMSETTING.DEFAULT)
            self.int.setHeuristics(scip.SCIP_PARAMSETTING.DEFAULT)

        # Attempt to solve the problem.
        with self._header(), self._stopwatch():
            self.int.optimize()

        # Retrieve primals.
        primals = {}
        if self.ext.options.primals is not False:
            for picosVar in self.ext.variables.values():
                value = []
                for localIndex in range(picosVar.dim):
                    scipVar = self._scipVar[picosVar.id_at(localIndex)]
                    value.append(self.int.getVal(scipVar))

                primals[picosVar] = value

        # Retrieve duals for LPs.
        duals = {}
        if self.ext.options.duals is not False and isLP:
            for picosConstraint in self.ext.constraints.values():
                assert isinstance(picosConstraint, AffineConstraint)

                # Retrieve dual value for constraint.
                scipDuals = []
                for scipCon in self._scipCons[picosConstraint]:
                    scipDuals.append(self.int.getDualsolLinear(scipCon))
                picosDual = cvxopt.matrix(scipDuals, picosConstraint.size)

                # Flip sign based on constraint relation.
                if picosConstraint.is_decreasing():
                    picosDual = -picosDual

                # Flip sign based on objective sense.
                if picosDual and self.ext.no.direction == "min":
                    picosDual = -picosDual

                duals[picosConstraint] = picosDual

        # # Retrieve objective value.
        # objectiveValue = self.int.getObjVal()
        #
        # # HACK: Add back the constant objective term; see above.
        # picosObjective = self.ext.no.function
        # picosObjectiveAffinePart = picosObjective.aff \
        #     if isinstance(picosObjective, QuadExp) else picosObjective
        # if picosObjectiveAffinePart.constant:
        #     objectiveValue += picosObjectiveAffinePart.constant[0]

        status = self.int.getStatus()
        if   status == "optimal":
            primalStatus   = SS_OPTIMAL
            dualStatus     = SS_OPTIMAL
            problemStatus  = PS_FEASIBLE
        elif status == "timelimit":
            primalStatus   = SS_PREMATURE
            dualStatus     = SS_PREMATURE
            problemStatus  = PS_UNKNOWN
        elif status == "infeasible":
            primalStatus   = SS_INFEASIBLE
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_INFEASIBLE
        elif status == "unbounded":
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_INFEASIBLE
            problemStatus  = PS_UNBOUNDED
        else:
            primalStatus   = SS_UNKNOWN
            dualStatus     = SS_UNKNOWN
            problemStatus  = PS_UNKNOWN

        return self._make_solution(
            primals, duals, primalStatus, dualStatus, problemStatus)


# --------------------------------------
__all__ = api_end(_API_START, globals())
